/*
 * Copyright (C) 2002-2003 Peter J Jones (pjones@pmade.org)
 * All Rights Reserved
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of the Author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "httpd.h"
#include "http_core.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_log.h"
#include "http_request.h"
#include "apr_tables.h"
#include "apr_strings.h"
#include "apr_file_info.h"

/* apache 2.2 compat */
#define regex_t ap_regex_t
#define REG_EXTENDED AP_REG_EXTENDED

// for getpid()
#include <sys/types.h>
#include <unistd.h>

/* ######################################################################## */
module AP_MODULE_DECLARE_DATA variety_module;
/* ######################################################################## */
struct variety_config {
    int variety_on;
    int use_cookies;
    regex_t *re_include;
    regex_t *re_exclude;
    apr_size_t directory_max;
};
/* ######################################################################## */
struct variety_cookie_it {
    request_rec *r;
    const char *filename;
};
/* ######################################################################## */
static int variety_fixups (request_rec *r);
static int variety_redirect_handler (request_rec *r);
apr_array_header_t* variety_list_directory (request_rec *r, struct variety_config *c, const char *dir);
static int variety_select_file (request_rec *r, apr_array_header_t *matched_files);
static const char* variety_get_cookie (request_rec *r);
static int variety_cookie_iterate (void *out, const char *key, const char *value);
static void variety_set_cookie (request_rec *r, const char *filename);
static void* variety_create_config (apr_pool_t *p);
static void* variety_create_dir_config (apr_pool_t *p, char *dirspec); 
static const char* variety_cfg_on (cmd_parms *parms, void *mconfig, int on);
static const char* variety_cfg_cookies (cmd_parms *parms, void *mconfig, int on);
static const char* variety_cfg_match (cmd_parms *parms, void *mconfig, const char *w);
static const char* variety_cfg_exclude (cmd_parms *parms, void *mconfig, const char *w);
static const char* variety_cfg_dirmax (cmd_parms *parms, void *mconfig, const char *w);
static void variety_register_hooks(apr_pool_t *p);
/* ######################################################################## */
static int variety_fixups (request_rec *r) {
    struct variety_config *c;
    request_rec *other_request;
    apr_array_header_t *matched_files;
    char *dirname, *pos;
    char **element;
    int i, urilen, pinfolen, filelen;

/*ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "variety: running");*/

    /* reasons to not handle this request */
    if (r->proxyreq != NULL || r->main != NULL) {
	return DECLINED;
    }

    /*
     * we need to make sure that we have not already run for this request
     * chain. If we run more than once it could cause an infinite loop.
     */
    for (other_request = r->prev; other_request != NULL && other_request->handler != NULL; other_request = other_request->prev) {
	if (strcmp(other_request->handler, "variety_redirect_handler") == 0) return DECLINED;
    }

    /* check to see if we were configured and turned on */
    if ( (c = (struct variety_config*)ap_get_module_config(r->per_dir_config, &variety_module)) == NULL || !c->variety_on) {
	return DECLINED;
    }

    /* figure out what directory to open and scan */
    dirname = apr_pstrdup(r->pool, r->filename);
    pos = dirname + strlen(dirname) - 1;
    while (pos != dirname && *pos != '/') --pos;
    *pos = '\0';

    /* make sure that the URI and the filename line up */
    filelen = strlen(pos+1);
    urilen = strlen(r->uri);
    pinfolen = r->path_info == NULL ? 0 : strlen(r->path_info);

    if (strncmp(pos+1, r->uri + (urilen - (pinfolen + filelen)), filelen) != 0) {
	ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r, "variety: URI '%s' and directory '%s' don't line up", r->uri, r->filename);
	return DECLINED;
    }

    if ( (matched_files = variety_list_directory(r, c, dirname)) == NULL) return HTTP_FORBIDDEN;
    element = (char**)matched_files->elts;

    if (matched_files->nelts == 0) {
	ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r, "variety: no matching files in %s", dirname);
	return DECLINED;
    }

    i = variety_select_file(r, matched_files);
    if (c->use_cookies) variety_set_cookie(r, element[i]);

    r->handler = "variety_redirect_handler";
    r->filename = apr_pstrcat(r->pool, "redirect:", apr_pstrndup(r->pool, r->uri, (urilen - (pinfolen + filelen))),
	    element[i], r->path_info == NULL ? "" : r->path_info, r->args ? "?" : NULL, r->args, NULL);

    return OK;
}
/* ######################################################################## */
static int variety_redirect_handler (request_rec *r) {
    if (strcmp(r->handler, "variety_redirect_handler") != 0) return DECLINED;
    if (strncmp(r->filename, "redirect:", 9) != 0) return DECLINED;

    ap_internal_redirect(r->filename+9, r);
    return OK;
}
/* ######################################################################## */
static int variety_select_file (request_rec *r, apr_array_header_t *matched_files) {
    int total_entries = matched_files->nelts;
    if (total_entries == 1) return 0;
    return ((int)(((double)(rand() % RAND_MAX) / RAND_MAX) * total_entries + 1) % total_entries);
}
/* ######################################################################## */
apr_array_header_t* variety_list_directory (request_rec *r, struct variety_config *c, const char *dir) {
    apr_array_header_t *rv = apr_array_make(r->pool, 0, sizeof(char*));
    apr_dir_t *directory_handle;
    apr_finfo_t current_file;
    apr_size_t array_size=0;
    const char *cookie_file = NULL;

    /* get the last file picked from a cookie */
    if (c->use_cookies) cookie_file = variety_get_cookie(r);

    if (apr_dir_open(&directory_handle, dir, r->pool) != APR_SUCCESS) {
	ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r, "variety: failed to open directory %s", dir);
	return NULL;
    }

    while (apr_dir_read(&current_file, APR_FINFO_TYPE | APR_FINFO_NAME, directory_handle) == APR_SUCCESS) {
	char **new_element;

	if (current_file.filetype != APR_REG && current_file.filetype != APR_LNK) continue;

	if (cookie_file && strcmp(cookie_file, current_file.name) == 0) {
	    /* this file was served last time and stored in a cookie */
	    cookie_file = NULL;
	    continue;
	}

	if (c->re_include != NULL) {
	    /* should we include this item? */
	    if (ap_regexec(c->re_include, current_file.name, 0, NULL, 0) != 0) continue;
	}

	if (c->re_exclude != NULL) {
	    /* should we exclude this item? */
	    if (ap_regexec(c->re_exclude, current_file.name, 0, NULL, 0) == 0) continue;
	}

	/*
	 * We are going to keep this file. Make sure that we don't exceed
	 * the size limit of the array. This is a nice way to make this code
	 * smaller and faster yet still under the control of the system
	 * administrator.
	 */
	array_size += strlen(current_file.name);

	if (c->directory_max && array_size > c->directory_max) {
	    ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r, "variety: directory max hit for %s", dir);
	    break;
	}

	new_element = apr_array_push(rv);
	*new_element = apr_pstrdup(r->pool, current_file.name);
    }

    apr_dir_close(directory_handle);
    return rv;
}
/* ######################################################################## */
static const char* variety_get_cookie (request_rec *r) {
    struct variety_cookie_it cookie = {r, NULL};
    apr_table_do(variety_cookie_iterate, &cookie, r->headers_in, "Cookie", NULL);
    return cookie.filename;
}
/* ######################################################################## */
static int variety_cookie_iterate (void *out, const char *key, const char *value) {
    struct variety_cookie_it *cookie = (struct variety_cookie_it*)out;
    const char *where, *end;

    if ( (where = ap_strstr(value, "variety-lf=")) == NULL) return 1;

    where += 11;
    end = where;

    while (end != NULL && *end != '\0' && *end != ';') ++end;
    if (end == where) return 1;

    cookie->filename = apr_pstrmemdup(cookie->r->pool, where, end - where);
    return 0;
}
/* ######################################################################## */
static void variety_set_cookie (request_rec *r, const char *filename) {
    /* 
     * set a cookie, use err_headers_out because headers_out does not go to
     * the interal redirect, and thus won't make it to the browser.
     */
    char *cookie_str = apr_pstrcat(r->pool, "variety-lf=", filename, ";", NULL);
    apr_table_addn(r->err_headers_out, "Set-Cookie", cookie_str);
}
/* ######################################################################## */
static void* variety_create_config (apr_pool_t *p) {
    struct variety_config *c = (struct variety_config*)apr_pcalloc(p, sizeof(*c));

    c->variety_on = 0;
    c->use_cookies = 0;
    c->re_include = NULL;
    c->re_exclude = NULL;
    c->directory_max = 1024;

    return c;
}
/* ######################################################################## */
static void* variety_create_dir_config (apr_pool_t *p, char *dirspec) {
    return variety_create_config(p);
}
/* ######################################################################## */
static const char* variety_cfg_on (cmd_parms *parms, void *mconfig, int on) {
    struct variety_config *c = (struct variety_config*)mconfig;
    c->variety_on = on == 1;

    return NULL;
}
/* ######################################################################## */
static const char* variety_cfg_cookies (cmd_parms *parms, void *mconfig, int on) {
    struct variety_config *c = (struct variety_config*)mconfig;
    c->use_cookies = on == 1;

    return NULL;
}
/* ######################################################################## */
static const char* variety_cfg_match (cmd_parms *parms, void *mconfig, const char *w) {
    struct variety_config *c = (struct variety_config*)mconfig;
    if (c->re_include != NULL) return "VarietyMatch has already been set for this location";

    if ( (c->re_include = ap_pregcomp(parms->pool, w, REG_EXTENDED)) == NULL) {
	return apr_pstrcat(parms->pool, "VarietyMatch: failed to compile regular expression '", w, "'", NULL);
    }

    return NULL;
}
/* ######################################################################## */
static const char* variety_cfg_exclude (cmd_parms *parms, void *mconfig, const char *w) {
    struct variety_config *c = (struct variety_config*)mconfig;
    if (c->re_exclude != NULL) return "VarietyExclude has already been set for this location";

    if ( (c->re_exclude = ap_pregcomp(parms->pool, w, REG_EXTENDED)) == NULL) {
	return apr_pstrcat(parms->pool, "VarietyExclude: failed to compile regular expression '", w, "'", NULL);
    }

    return NULL;
}
/* ######################################################################## */
static const char* variety_cfg_dirmax (cmd_parms *parms, void *mconfig, const char *w) {
    struct variety_config *c = (struct variety_config*)mconfig;
    if (c->directory_max != 1024) return "VarietyDirMax has already been set for this location";

    c->directory_max = atoi(w);
    return NULL;
}

/* ######################################################################## */
static int variety_init_handler(apr_pool_t *p, apr_pool_t *plog,
                                apr_pool_t *ptemp, server_rec *s)
{
    ap_add_version_component(p, "mod_variety/0.2.0");
    return OK;
}

/* ######################################################################## */
static void variety_register_hooks (apr_pool_t *p) {
    ap_hook_post_config(variety_init_handler, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_handler(variety_redirect_handler, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_fixups(variety_fixups, NULL, NULL, APR_HOOK_FIRST);

    /* seed the (fake) random number generator */
    srand((unsigned)(getpid()));
}
/* ######################################################################## */
static const command_rec variety_cmds[] = {
    AP_INIT_FLAG("Variety",         variety_cfg_on,         NULL, OR_LIMIT, "Turn mod_variety On of Off"),
    AP_INIT_FLAG("VarietyCookies",  variety_cfg_cookies,    NULL, OR_LIMIT, "Turn mod_variety Cookies On or Off"),
    AP_INIT_TAKE1("VarietyMatch",   variety_cfg_match,	    NULL, OR_LIMIT, "Include files that match the given regular expression"),
    AP_INIT_TAKE1("VarietyExclude", variety_cfg_exclude,    NULL, OR_LIMIT, "Exclude files that match the given regular expression"),
    AP_INIT_TAKE1("VarietyDirMax",  variety_cfg_dirmax,	    NULL, OR_LIMIT, "Limit the size of possible matches"),
    {NULL}
};
/* ######################################################################## */
module AP_MODULE_DECLARE_DATA variety_module = {
    STANDARD20_MODULE_STUFF, 
    variety_create_dir_config,
    NULL,                  
    NULL,
    NULL,                  
    variety_cmds,
    variety_register_hooks
};
/* ######################################################################## */
